import csv
from pathlib import Path
from typing import Dict, FrozenSet, Iterable, NamedTuple, NewType
from collections.abc import Set

from .paths import UD2UM_FILE


Form = NewType("Form", str)
Lemma = NewType("Lemma", str)
UmTag = NewType("UmTag", str)
UdFeat = NewType("UdFeat", str)
UmFeat = NewType("UmFeat", str)
UmFeats = FrozenSet[UmFeat]


class UniMorphTriple(NamedTuple):
    form: Form
    lemma: Lemma
    feats: Set[UmFeat]


class UdTag(Set):
    def __init__(self, tag: str) -> None:
        self.data = set(tag.split("|"))

    def __contains__(self, value):
        return value in self.data

    def __iter__(self):
        yield from self.data

    def __len__(self):
        return len(self.data)


def ud_iterator(file: Path) -> Iterable[str]:
    with open(file, encoding="utf-8") as f:
        yield from (line.strip() for line in f)


class CoNLLRow(NamedTuple):
    id: str
    form: Form
    lemma: Lemma
    upostag: str
    xpostag: str
    feats: str
    head: str
    deprel: str
    deps: str
    misc: str

    @classmethod
    def make(cls, string: str) -> "CoNLLRow":
        return cls._make(string.split("\t"))


def is_conll_useless(line: str) -> bool:
    if line == "":
        return True
    if line.startswith("#"):
        return True
    return False


def _ud2um_mapping() -> Dict[UdFeat, UmFeat]:
    ud2um = {}
    with open(UD2UM_FILE, encoding="utf-8") as f:
        reader = csv.DictReader(f, delimiter="\t")
        for row in reader:
            ud = UdFeat(row["UD"])
            um = UmFeat(row["UniMorph"] or "_")
            ud2um[ud] = um
    print(ud2um)
    return ud2um


ud2um_mapping = _ud2um_mapping()



"""
Convert Universal Dependencies morphology annotations to UniMorph.
"""

from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Iterable, List, Optional, Tuple, Union

from termcolor import cprint

from .languages import languages, LanguageCoding, get_lang
from .paths import FileGetter, output_filepath
from .translator import Translator, translators
from .um_reader import unimorph
from .utils import CoNLLRow, is_conll_useless, ud_iterator


class EvaluationInstance:
    def __init__(
        self,
        language: LanguageCoding,
        clever=False,
        replace_feats=False,
        print_good=False,
    ) -> None:
        self.language = language
        self.um_file, self.ud_files = FileGetter.get(language, replace_feats)
        translator_class = translators.get(language, Translator)
        self.translator = translator_class(clever, replace_feats)
        self.print_good = print_good
        # print(self.translator)

        if not replace_feats:
            self.tags, self.lemmas = unimorph(self.um_file)

    def translate(
        self, source: str, output_all=False
    ) -> Optional[Union[CoNLLRow, str]]:
        if is_conll_useless(source):
            if output_all:
                return source
            else:
                return None
        record = CoNLLRow.make(source)
        updated = self.translator.translate(record)
        if output_all:
            return "\t".join(updated)
        else:
            return updated

    def evaluate(self) -> None:
        scores = []
        for file in self.ud_files:
            score = self._evaluate(file)
            scores.append(score)
            print(file.name, score)
        # Calculate recall.
        good_counts, counts = zip(*scores)
        print(
            f"Average for {self.language.name}:",
            sum(good_counts) / (sum(counts) or 1) * 100,
        )

    def convert(self) -> None:
        file: Path
        assert self.ud_files
        for file in self.ud_files:
            print(file)
            lines: Iterable[str] = ud_iterator(file)
            translations = [self.translate(line, output_all=True) for line in lines]
            with open(output_filepath(file), "w", encoding="utf-8") as f:
                for line in translations:
                    print(line, file=f)

    def score_translation(self, t: CoNLLRow) -> Tuple[int, int]:
        good_count = bad_count = count = 0
        try:
            tag = t.misc
            assert t.form in self.tags
            token_bundle = set(tag.split(";"))
            type_bundles = self.tags[t.form]
            assert t.lemma in self.lemmas[t.form]
            self.translator.lgspec_assert(t, token_bundle)
            if token_bundle in type_bundles:
                good_count += 1
                if self.print_good:
                    cprint(
                        f"{(t.form):20}\t{(';'.join(sorted(token_bundle))):20}\t{str([';'.join(sorted(tags)) for tags in type_bundles]):40}",
                        "green",
                    )
            else:
                cprint(
                    f"{(t.form):20}\t{(';'.join(sorted(token_bundle))):20}\t{str([';'.join(sorted(tags)) for tags in type_bundles]):40}",
                    "red",
                )
                bad_count += 1
            count += 1
        except AssertionError:
            pass
        assert good_count + bad_count == count
        return good_count, count

    def recall(self, translations: List[CoNLLRow]) -> Tuple[int, int]:
        good_count = count = 0
        for translation in translations:
            gc, c = self.score_translation(translation)
            good_count += gc
            count += c
        return good_count, count

    def _evaluate(self, file: Path) -> Tuple[int, int]:
        lines: Iterable[str] = ud_iterator(file)
        maybe_translations: List[Optional[Union[CoNLLRow, str]]] = [
            self.translate(line) for line in lines
        ]
        translations: List[CoNLLRow] = [
            t for t in maybe_translations if isinstance(t, CoNLLRow)
        ]
        recall = self.recall(translations)
        return recall


class FileConverter(EvaluationInstance):
    """docstring for FileConverter"""

    def __init__(self, file: Path, language: LanguageCoding, clever: bool) -> None:
        super(FileConverter, self).__init__(language, clever, replace_feats=True)
        self.ud_files = [file]


def parse_args() -> Namespace:
    parser = ArgumentParser(__doc__)
    subparsers = parser.add_subparsers(dest="command", help="sub-command help")

    # create the parser for the "a" command
    replicate = subparsers.add_parser(
        "replicate", help="replicate experiments from McCarthy et al. (2018)"
    )
    # parser_a.add_argument('bar', type=int, help='bar help')

    # create the parser for the "b" command
    evaluate = subparsers.add_parser("evaluate", help="evaluate a Translator class")
    evaluate.add_argument("-b", "--basic", action="store_true", help="dumb conversion?")
    evaluate.add_argument(
        "-p",
        "--print_good",
        action="store_true",
        help="Print good during evaluation, or only bad?",
    )
    evaluate.add_argument(
        "-l",
        "--langs",
        nargs="+",
        required=True,
        help='languages to convert (e.g. "da eu sp")',
    )

    # create the parser for the "b" command
    convert = subparsers.add_parser("convert", help="convert UD files to UniMorph")
    convert.add_argument("-b", "--basic", action="store_true", help="dumb conversion?")
    convert.add_argument(
        "--ud",
        type=Path,
        help="UD file; warning: this will not use the clever converter",
    )
    convert.add_argument(
        "-l", "--langs", nargs="+", help='languages to convert (e.g. "da eu sp")'
    )
    return parser.parse_args()


def replicate() -> None:
    for language in languages:
        cprint(language.name, attrs={"bold"})
        for clever in [False, True]:
            print("Clever? ", clever)
            instance: EvaluationInstance = EvaluationInstance(language, clever)
            instance.evaluate()


def evaluate(args: Namespace) -> None:
    for language_ in args.langs:
        language = get_lang(language_)
        cprint(language.name, attrs={"bold"})

        clever = not args.basic
        instance: EvaluationInstance = EvaluationInstance(
            language, clever, print_good=args.print_good
        )
        instance.evaluate()


def convert(args: Namespace) -> None:
    for language_ in args.langs:
        language = get_lang(language_)
        cprint(language.name, attrs={"bold"})

        clever = not args.basic
        instance: EvaluationInstance = EvaluationInstance(
            language, clever, replace_feats=True
        )
        instance.convert()


def convert_file(args: Namespace) -> None:
    assert not args.langs or len(args.langs) == 1
    if args.langs:
        try:
            [language_] = args.langs
            language = get_lang(language_)
        except KeyError:
            cprint(f"Warning: no clever converter exists for {args.ud}", "cyan")
            language = LanguageCoding(None, None, "")
    else:  # args.langs == None
        cprint(f"Warning: no clever converter exists for {args.ud}", "cyan")
        language = LanguageCoding(None, None, "")
    cprint(language.name, attrs={"bold"})
    clever = not args.basic

    instance: FileConverter = FileConverter(args.ud, language, clever)
    instance.convert()


def main() -> None:
    args = parse_args()
    print(args)
    if args.command == "replicate":
        replicate()
    elif args.command == "evaluate":
        evaluate(args)
    elif args.command == "convert":
        if args.ud:
            convert_file(args)
        else:
            convert(args)
    else:
        raise ValueError


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import List, Tuple

import pkg_resources

from .languages import LanguageCoding

# These must be pathlib Path objects.
# Easy way: Path('whatever/my/path/is')
_ROOT = Path("/Users/arya/Desktop/UD_UM") / "data" / "raw"
UM_FOLDER = _ROOT / "UM"
UD_FOLDER = _ROOT / "UD"
UD2UM_FILE = pkg_resources.resource_filename('ud_compatibility', 'UD-UniMorph.tsv')


class FileGetter:
    @staticmethod
    def get(language: LanguageCoding, convert=False) -> Tuple[Path, List[Path]]:
        um_file = UM_FOLDER / f"{language.um}-master" / f"{language.um}"
        lang_folder = UD_FOLDER / f"UD_{language.name}-master"
        print(lang_folder)
        ud_files = list(lang_folder.glob(f"{language.ud}-ud-*.conllu"))
        FileGetter._check_inputs(um_file, ud_files, convert)
        return um_file, ud_files

    @staticmethod
    def _check_inputs(um_file: Path, ud_files: List[Path], skip_um: bool) -> None:
        if not skip_um:
            FileGetter._validate_file(um_file)
        if not ud_files:
            print(f"No useful things in {UD_FOLDER}")
        for file in ud_files:
            FileGetter._validate_file(file)

    @staticmethod
    def _validate_file(file: Path) -> None:
        assert file.is_file(), file


def output_filepath(conllu: Path) -> Path:
    return conllu.with_name(conllu.name.replace("-ud-", "-um-"))



import re
from collections import defaultdict
from typing import List, Set

from .languages import languages
from .utils import CoNLLRow, UdFeat, UdTag, UmFeat, UmTag, ud2um_mapping

EMPTY_FEAT = UmFeat("_")


def handle_arguments(ud: UdTag) -> List[UmFeat]:
    def handle_argument(parts):
        parts = str(parts)
        if "[psed]" in parts or "[gram]" in parts:
            return "_"

        if "[erg]" in parts:
            kind = "ER"
        elif "[dat]" in parts:
            kind = "DA"
        elif "[abs]" in parts:
            kind = "AB"
        else:
            print(parts)
            raise AssertionError

        if "=Plur" in parts:
            number = "P"
        elif "=Sing" in parts:
            number = "S"
        elif "=Dual" in parts:
            number = "D"
        else:
            assert "Number" not in str(parts)
            number = ""

        if "=1" in parts:
            person = "1"
        elif "=2" in parts:
            person = "2"
        elif "=3" in parts:
            person = "3"
        else:
            assert "Person" not in str(parts)
            person = ""
        return f"ARG{kind}{person}{number}"

    arg_parts = [p for p in ud if "[" in p and "[psor]" not in p]
    if not arg_parts:
        return [EMPTY_FEAT]
    arg_re = re.compile(r"\[(.*?)\]")
    tags = {arg_re.search(p).group(1) for p in arg_parts}  # type: ignore
    contributions = []
    for tag in tags:
        arg = handle_argument([p for p in arg_parts if tag in p])
        if arg:
            contributions.append(arg)
    return contributions


def handle_possession(ud_tag: UdTag) -> UmFeat:
    psor_parts = [p for p in ud_tag if "[psor]" in p]
    if not psor_parts:
        return EMPTY_FEAT

    if "None" in str(psor_parts):
        return UmFeat("PSSD")

    try:
        assert len(psor_parts) <= 2
    except AssertionError:
        print(psor_parts)
        raise

    if "Number[psor]=Plur" in psor_parts:
        number = "P"
    elif "Number[psor]=Sing" in psor_parts:
        number = "S"
    elif "Number[psor]=Dual" in psor_parts:
        number = "D"
    else:
        assert "Number" not in str(psor_parts)
        number = ""

    if "Person[psor]=1" in psor_parts:
        person = "1"
    elif "Person[psor]=2" in psor_parts:
        person = "2"
    elif "Person[psor]=3" in psor_parts:
        person = "3"
    else:
        assert "Person" not in str(psor_parts)
        person = ""

    return UmFeat(f"PSS{person}{number}")


def process_tag(part: UdFeat) -> UmFeat:
    try:
        um_part = ud2um_mapping[part]
    except KeyError:
        # print("Couldn't find", part)
        return EMPTY_FEAT
    #     if part != "_":
    # if part in UD_valid_tags:
    #     self.no_match_tags.add(part)
    # else:
    #     self.invalid_tags.add(part)
    else:
        return um_part


def ud2um(ud_tag: UdTag) -> UmTag:
    um_tag: List[UmFeat] = []
    possession = handle_possession(ud_tag)
    um_tag.append(possession)
    arguments = handle_arguments(ud_tag)
    um_tag.extend(arguments)

    for part in ud_tag:
        if "," not in part:
            tag = process_tag(part)
            um_tag.append(tag)
        else:
            key, vals = part.split("=")
            vals = vals.split(",")
            all_parts = []
            for val in vals:
                tag = process_tag(UdFeat(f"{key}={val}"))
                all_parts.append(tag)
            all_parts = [p for p in all_parts if p != "_"]
            um_tag.append(
                UmFeat(f"{{{'/'.join(all_parts)}}}") if all_parts else EMPTY_FEAT
            )
    um_tag = [f for f in um_tag if str(f) != "_"] or [EMPTY_FEAT]
    # print(um_tag)
    return UmTag(";".join(um_tag))


class Translator:
    def __init__(self, clever, replace_feats) -> None:
        self.clever = clever
        self.replace_feats = replace_feats

    def translate(self, record: CoNLLRow) -> CoNLLRow:
        ud_tag: UdTag = UdTag(f"{record.upostag}|{record.feats}")
        # print(ud_tag)
        um_tag: UmTag = self.basic_convert(ud_tag)

        if self.clever:
            um_tag = self.lgspec_modify(record, um_tag)

        if self.replace_feats:
            updated = record._replace(feats=um_tag)
        else:
            updated = record._replace(misc=um_tag)
        return updated

    def basic_convert(self, ud_tag: UdTag) -> UmTag:
        return ud2um(ud_tag)

    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        """Override me."""
        pass

    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        return um


class BasqueTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        assert "V" in tags or "V.PTCP" in tags or "V.CVB" in tags or "V.MSDR" in tags


class BulgarianTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        assert "N" not in tags

    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        if "V" in tags:
            tags.discard("FIN")
            tags.discard("ACT")
            tags.discard("IPFV")
            tags.discard("PFV")
        if "V.PTCP" in tags:
            tags.discard("ADJ")
            tags.discard("V")
            if "PRS" in tags:
                tags.discard("IPFV")
            if "PASS" not in tags:
                tags.add("ACT")
            if "PFV" in tags:
                tags.discard("PFV")
                tags.add("PST")
        if "RL" in tags:
            tags.remove("RL")
            tags.add("SPRL")
        return UmTag(";".join(tags))


class CatalanTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        assert "V" in tags or "V.PTCP" in tags or "V.CVB" in tags or "V.MSDR" in tags

    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        # if "N" in tags:
        #     for gender in {"MASC", "FEM", "NEUT", "{MASC/NEUT}"}:
        #         tags.discard(gender)
        if "V.PTCP" in tags:
            tags.discard("V")
        if "V" in tags:
            tags.discard("FIN")
            if "PST" in tags:
                tags.add("PFV")
            if "IPFV" in tags:
                tags.add("PST")
            if "COND" in tags:
                tags.discard("PRS")
            if "IMP" in tags:
                tags.add("POS")
                tags.discard("PRS")
        if tags == set("MASC;PST;SG;V;V.PTCP".split(";")):
            tags = {"PST", "V.PTCP"}
        if tags == set("V;V.MSDR".split(";")):
            tags = set("PRS;V.PTCP".split(";"))
        if {"IPFV", "SBJV"} < tags:
            tags.remove("IPFV")
        return UmTag(";".join(tags))


class CzechTranslator(Translator):
    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        if "ADJ" in tags:
            tags.discard("POS")
        if "N" in tags:
            tags.discard("POS")
            tags.discard("INAN")
            tags.discard("ANIM")
            for gender in {"MASC", "FEM", "NEUT", "{MASC/NEUT}"}:
                if gender in tags:
                    tags.remove(gender)
        if "V" in tags:
            tags.discard("ACT")
            tags.discard("FIN")
            tags.discard("POS")
        return UmTag(";".join(tags))


class DanishTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        assert "ADJ" not in tags

    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        if "V" in tags:
            tags.discard("FIN")
        return UmTag(";".join(tags))


class DutchTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        assert "N" not in tags


class EnglishTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        assert "V" in tags or "V.PTCP" in tags or "V.CVB" in tags or "V.MSDR" in tags

    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        tags.discard("FIN")
        tags.discard("IND")
        if tags == {"V", "PRS"}:
            tags = {"V", "NFIN"}
        if tags == {"V", "V.MSDR"}:
            tags = {"PRS", "V", "V.PTCP"}
        if "V" in tags:
            tags.discard("PASS")
        return UmTag(";".join(tags))


class EstonianTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        assert "ADJ" not in tags


class FinnishTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        assert tags != set("ADJ;GEN;SG".split(";"))
        assert tags != set("ADJ;GEN;PL".split(";"))
        assert "0" not in tags

    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        if "V" in tags:
            if "FIN" in tags:
                tags.remove("FIN")
                tags.add("POS")
        if tags == set("NOM;PASS;SG;V;V.PTCP".split(";")):
            tags = set("PASS;PST;V.PTCP".split(";"))
        if tags == set("ACT;NOM;SG;V;V.PTCP".split(";")):
            tags.remove("NOM")
            tags.remove("SG")
            tags.remove("V")
            tags.add("PST")
        if tags == set("ACT;NFIN;SG;V".split(";")):
            tags.remove("ACT")
            tags.remove("SG")
        return UmTag(";".join(tags))


class FrenchTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        assert "V" in tags or "V.PTCP" in tags or "V.CVB" in tags or "V.MSDR" in tags

    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        # if "N" in tags:
        #     for gender in {"MASC", "FEM", "NEUT", "{MASC/NEUT}"}:
        #         tags.discard(gender)
        if "V.PTCP" in tags:
            tags.discard("V")
            tags.discard("MASC")
            tags.discard("FEM")
            tags.discard("SG")
            tags.discard("PL")
        if "V" in tags:
            tags.discard("FIN")
            if "PST" in tags:
                tags.add("PFV")
            if "IPFV" in tags:
                tags.add("PST")
            if "COND" in tags:
                tags.discard("PRS")
            if "IMP" in tags:
                tags.add("POS")
                tags.discard("PRS")
        if tags == set("MASC;PST;SG;V;V.PTCP".split(";")):
            tags = {"PST", "V.PTCP"}
        if tags == set("V;V.MSDR".split(";")):
            tags = set("PRS;V.CVB".split(";"))
        return UmTag(";".join(tags))


class GermanTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        assert (
            "V" in tags
            or "N" in tags
            or "V.PTCP" in tags
            or "V.CVB" in tags
            or "V.MSDR" in tags
        )
        # assert tags != {"N"}
        # assert tags != {"V"}

    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        for gender in {"MASC", "FEM", "NEUT", "{MASC/NEUT}"}:
            if gender in tags:
                tags.remove(gender)
        tags.discard("FIN")
        if tags == {"V", "V.PTCP"}:
            tags = {"PST", "V.PTCP"}
        return UmTag(";".join(tags))


# translators[get_lang("German")] = GermanTranslator


class HebrewTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        assert "ADJ" not in tags

    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        for gender in {"MASC", "FEM", "NEUT", "{FEM/MASC}"}:
            if gender in tags:
                tags.remove(gender)
        if "V.PTCP" in tags:
            tags.discard("V")
        if "V" in tags:
            tags.discard("ACT")
            tags.discard("POS")
        if "N" in tags:
            if len(tags) == 2:
                if "SG" in tags or "PL" in tags:
                    tags.add("NDEF")

        return UmTag(";".join(tags))


class HungarianTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        assert "ADJ" not in tags

    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        tags.discard("FIN")
        tags.discard("ACT")
        if "NFIN" in tags:
            tags.discard("PRS")
        # if "IN+ABL" in tags:
        #     tags.discard("IN+ABL")
        #     tags.add("ON+ABL")
        # if "IN+ESS" in tags:
        #     tags.discard("IN+ESS")
        #     tags.add("ON+ESS")
        # if "IN+ALL" in tags:
        #     tags.discard("IN+ALL")
        #     tags.add("ON+ALL")
        return UmTag(";".join(tags))


class ItalianTranslator(Translator):
    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        # if "N" in tags:
        #     for gender in {"MASC", "FEM", "NEUT", "{MASC/NEUT}"}:
        #         tags.discard(gender)
        if "V.PTCP" in tags:
            tags.discard("V")
            tags.discard("MASC")
            tags.discard("FEM")
            tags.discard("SG")
            tags.discard("PL")
        if "V" in tags:
            tags.discard("FIN")
            if "PST" in tags:
                tags.add("PFV")
            if "IPFV" in tags:
                tags.add("PST")
            if "COND" in tags:
                tags.discard("PRS")
        if tags == set("MASC;PST;SG;V;V.PTCP".split(";")):
            tags = {"PST", "V.PTCP"}
        if tags == set("V;V.MSDR".split(";")):
            tags = set("PRS;V.CVB".split(";"))
        return UmTag(";".join(tags))


class LatinTranslator(Translator):
    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        if "N" in tags:
            for gender in {"MASC", "FEM", "NEUT", "{MASC/NEUT}"}:
                tags.discard(gender)
        if "V" in tags:
            tags.discard("ACT")
            tags.discard("FIN")
        return UmTag(";".join(tags))


class LatvianTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        assert tags != set("3;IND;PRS;V".split(";"))
        assert tags != set("3;IND;PST;V".split(";"))
        assert tags != set("3;IND;FUT;V".split(";"))
        assert "ESS" not in tags

    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        if "V" in tags:
            tags.discard("FH")
            tags.discard("POS")
            tags.discard("NEG")
            tags.discard("ACT")
            tags.discard("FIN")
            tags.discard("REFL")
        if "N" in tags:
            for gender in {"MASC", "FEM", "NEUT", "{MASC/NEUT}"}:
                tags.discard(gender)
        if tags == set("COND;V".split(";")):
            tags.add("PRS")
        return UmTag(";".join(tags))


class Norwegian_BokmaalTranslator(Translator):
    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        tags.discard("IND")
        tags.discard("FIN")
        return UmTag(";".join(tags))


class Norwegian_NynorskTranslator(Translator):
    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        tags.discard("IND")
        tags.discard("FIN")
        return UmTag(";".join(tags))


class PolishTranslator(Translator):
    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        if "N" in tags:
            for gender in {"MASC", "FEM", "NEUT"}:
                tags.discard(gender)
            tags.discard("INAN")
            tags.discard("HUM")
            tags.discard("NHUM")
        if "ADJ" in tags:
            if "PL" in tags:
                for gender in {"MASC", "FEM", "NEUT"}:
                    tags.discard(gender)
            tags.discard("HUM")
            tags.discard("INAN")
        if set("IPFV;NFIN;V".split(";")) == tags:
            tags.remove("IPFV")
        if set("PFV;NFIN;V".split(";")) == tags:
            tags.remove("PFV")
        if "V" in tags:
            tags.discard("ACT")
            tags.discard("FIN")
            tags.discard("IND")
            tags.discard("IPFV")
        return UmTag(";".join(tags))


class PortugueseTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        assert "V" in tags or "V.PTCP" in tags or "V.CVB" in tags or "V.MSDR" in tags

    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        if "V.PTCP" in tags:
            tags.discard("V")
            if "PASS" in tags:
                tags.remove("PASS")
                tags.add("PST")
            if "PRS" not in tags:
                tags.add("PST")
            if (cols.form.endswith("do") or cols.form.endswith("to")) and tags == set(
                "PST;V.PTCP".split(";")
            ):
                tags.add("MASC")
                tags.add("SG")
        if tags == {"V", "V.MSDR"}:
            tags = {"V.PTCP", "PRS"}
        if "V" in tags:
            tags.discard("FIN")
            if "PST" in tags and "IPFV" not in tags:
                tags.add("PFV")
            if "IPFV" in tags:
                tags.add("PST")
            if cols.form.endswith("ram") and tags == set("3;IND;PL;V".split(";")):
                tags.add("PFV")
                tags.add("PST")
            tags.discard("PASS")
            if "PST+PRF" in tags:
                tags.add("PST")
                tags.add("PRF")
                tags.remove("PST+PRF")
        return UmTag(";".join(tags))


class RomanianTranslator(Translator):
    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        if "V" in tags:
            tags.discard("FIN")
        if "{ACC/NOM}" in tags:
            tags.remove("{ACC/NOM}")
            tags.add("NOM/ACC")
        if "N" in tags:
            for gender in {"MASC", "FEM", "NEUT", "{MASC/NEUT}"}:
                tags.discard(gender)
        if {"V", "PST", "IND"} < tags:
            tags.add("PFV")
        return UmTag(";".join(tags))


class SlovenianTranslator(Translator):
    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        if "N" in tags:
            for gender in {"MASC", "FEM", "NEUT", "{MASC/NEUT}"}:
                tags.discard(gender)
            tags.discard("ANIM")
        return UmTag(";".join(tags))


class SpanishTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        assert "V" in tags or "V.PTCP" in tags or "V.CVB" in tags or "V.MSDR" in tags

    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        tags.discard("FIN")
        if "AUX" in tags:
            tags.remove("AUX")
            tags.add("V")
        if {"V", "V.PTCP"} < tags:
            tags.remove("V")
        if {"V", "V.MSDR"} == tags:
            tags = {"PRS", "V.CVB"}
        if "PST" in tags and "V.PTCP" not in tags:
            tags.add("PFV")
        if {"IND", "IPFV"} < tags:
            tags.add("PST")
        if "IMP" in tags:
            tags.add("POS")
        if {"IPFV", "SBJV"} < tags:
            tags.remove("IPFV")
            tags.add("PST")
            if cols.form[-2:] == "ra" or cols.form[-3:] == "ran":
                tags.add("LGSPEC1")
        return UmTag(";".join(tags))


# translators[get_lang("Spanish")] = SpanishTranslator


class SwedishTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        # assert "CMPR" not in tags
        pass

    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        tags.discard("FIN")
        if "ADJ" in tags:
            tags.discard("NOM")
            if "V.PTCP" in tags:
                tags.remove("V.PTCP")
                tags.discard("PST")
                tags.discard("PRS")
        if tags == {"ADJ", "PL"}:
            tags.add("INDF")
        if "N" in tags:
            tags.discard("MASC+FEM")
            tags.discard("NEUT")
        if tags == {"ACT", "SUP", "V"} or tags == {"PASS", "SUP", "V"}:
            tags.remove("SUP")
            tags.remove("V")
            tags.add("V.CVB")
        if tags == {"ADJ", "DEF", "SG"}:
            tags = {"ADJ", "DEF"}
        if {"IND", "PRS", "V"} == tags:
            tags.add("ACT")
        if {"ACT", "IMP", "V"} == tags:
            tags.remove("ACT")
        return UmTag(";".join(tags))


class TurkishTranslator(Translator):
    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        if "N" in tags:
            tags.discard("3")
        return UmTag(";".join(tags))

    # def lgspec_assert(self, cols, tags):
    #     assert "ESS" not in tags


class UkrainianTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        # assert "DET" not in tags
        pass

    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        if "N" in tags:
            for gender in {"MASC", "FEM", "NEUT", "{MASC/NEUT}"}:
                if gender in tags:
                    tags.remove(gender)
        if "N" in tags:
            tags.discard("INAN")
            tags.discard("ANIM")
        if tags >= {"NFIN", "V"}:
            tags.discard("IPFV")
            tags.discard("PFV")
        if "V" in tags:
            for tag in "FIN;IND;PFV;IPFV".split(";"):
                tags.discard(tag)
        if "V.CVB" in tags:
            tags.discard("V")
        return UmTag(";".join(tags))


class UrduTranslator(Translator):
    def lgspec_assert(self, cols: CoNLLRow, tags: Set[str]) -> None:
        # assert "PART" not in tags
        assert "ADJ" not in tags

    def lgspec_modify(self, cols: CoNLLRow, um: UmTag) -> UmTag:
        tags = set(um.split(";"))
        for gender in {"MASC", "FEM", "NEUT", "{MASC/NEUT}"}:
            if gender in tags:
                tags.remove(gender)
        if {"N", "SG"} <= tags or {"N", "PL"} <= tags or "PROPN" in tags:
            tags.discard("3")
        return UmTag(";".join(tags))


translators = dict()
for language in languages:
    name = language.name
    try:
        translators[language] = getattr(
            __import__(__name__), "".join(name.replace("-", "_")) + "Translator"
        )
    except (NameError, AttributeError):
        pass



from collections import namedtuple
from typing import List


class LanguageCoding(namedtuple("LanguageCoding", "ud um name")):
    __slots__ = ()


_languages = [
    # These are needed to provide paths to the datasets.
    #  - The two-letter helps identify the UD filename.
    #  - The three-letter helps identify the UniMorph folder and filename.
    #  - The full name identifies the UD folder and the Translator object.
    # (ISO 693-1 (usually), ISO 693-3, English name)
    ("ar", "ara", "Arabic"),
    ("eu", "eus", "Basque"),
    ("bg", "bul", "Bulgarian"),
    ("ca", "cat", "Catalan"),
    ("cs", "ces", "Czech"),
    ("da", "dan", "Danish"),
    ("nl", "nld", "Dutch"),
    ("en", "eng", "English"),
    ("fi", "fin", "Finnish"),
    ("fr", "fra", "French"),
    ("de", "deu", "German"),
    ("he", "heb", "Hebrew"),
    ("hi", "hin", "Hindi"),
    ("hu", "hun", "Hungarian"),
    ("ga", "gle", "Irish"),
    ("it", "ita", "Italian"),
    ("la", "lat", "Latin"),
    ("lv", "lav", "Latvian"),
    ("lt", "lit", "Lithuanian"),
    ("no_bokmaal", "nob", "Norwegian-Bokmaal"),
    ("no_nynorsk", "nno", "Norwegian-Nynorsk"),
    ("fa", "fas", "Persian"),
    ("pl", "pol", "Polish"),
    ("pt", "por", "Portuguese"),
    ("ro", "ron", "Romanian"),
    ("ru", "rus", "Russian"),
    ("sl", "slv", "Slovenian"),
    ("es", "spa", "Spanish"),
    ("sv", "swe", "Swedish"),
    ("tr", "tur", "Turkish"),
    ("uk", "ukr", "Ukrainian"),
    ("ur", "urd", "Urdu"),
    ("el", "gre", "Greek") # no converter available yet!
]


languages: List[LanguageCoding] = [LanguageCoding._make(l) for l in _languages]


def get_lang(needle: str) -> LanguageCoding:
    for language in languages:
        if needle == language.ud:
            return language
    else:
        raise KeyError(needle)



from collections import defaultdict
from pathlib import Path
from typing import Iterable, Tuple, Set, Dict


from .utils import Form, Lemma, UmFeat, UmFeats, UniMorphTriple


def _read_unimorph(fname: Path) -> Iterable[UniMorphTriple]:
    with open(fname, encoding="utf-8") as f:
        for line in f:
            if line.split():
                try:
                    lemma, inflected, features = line.strip().split("\t")
                except ValueError:
                    print("Line:", line.split())
                    raise
                features_typed = map(UmFeat, features.split(";"))
                yield UniMorphTriple(Form(inflected), Lemma(lemma), set(features_typed))


def _as_dict_of_sets(
    rows: Iterable[UniMorphTriple]
) -> Tuple[Dict[Form, Set[UmFeats]], Dict[Form, Set[Lemma]]]:
    tags: Dict[Form, Set[UmFeats]] = defaultdict(set)
    lemmas: Dict[Form, Set[Lemma]] = defaultdict(set)
    for form, lemma, tag in rows:
        tags[form].add(frozenset(tag))
        lemmas[form].add(lemma)
    return dict(tags), dict(lemmas)


def unimorph(fname: Path) -> Tuple[Dict[Form, Set[UmFeats]], Dict[Form, Set[Lemma]]]:
    return _as_dict_of_sets(_read_unimorph(fname))
